__all__ = ['run_script', 'Lens']

from typing import Callable, Final

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import InternalError, GenericError

from psuctl.firmware.psuctl import PsuctlFirmware
from psuctl.firmware.vitek import VitekFirmware
from psuctl.uconf import PsuctlUconf, ScriptInfo
from psuctl.types.unit import UnitType, UnitRef, Unit, IUnit
from psuctl.units.led import factory as led_unit_factory
from psuctl.units.psu import factory as psu_unit_factory
from psuctl.units.sh import sh_unit_factory, sh_ics_unit_factory
from psuctl.units.stepper import factory as stepper_unit_factory
from psuctl.units.stepper_vitek import factory as stepper_vitek_unit_factory

_UTYPE_TO_FIRMWARE_FACTORY_MAP = {
    'led': PsuctlFirmware,
    'psu': PsuctlFirmware,
    'stepper': PsuctlFirmware,
    'stepper-vitek': VitekFirmware,
    'sh': None,
    'sh-ics': None
}

UnitFactory = Callable[[PsuctlFirmware | VitekFirmware, UnitRef], IUnit | None]

_UTYPE_TO_UNIT_FACTORY_MAP: dict[UnitType, UnitFactory] = {
    'led': led_unit_factory,
    'psu': psu_unit_factory,
    'stepper': stepper_unit_factory,
    'stepper-vitek': stepper_vitek_unit_factory,
    'sh': sh_unit_factory,
    'sh-ics': sh_ics_unit_factory,
}


def run_script(endpoint: str, script_info: ScriptInfo, override_conf: dict[str, str], lock_path: str = '') -> None:
    if script_info.handler_utype not in _UTYPE_TO_FIRMWARE_FACTORY_MAP:
        eventlog.log_e(f'unsupported (handler-utype:{script_info.handler_utype})')
        raise InternalError

    unit_factory = _UTYPE_TO_UNIT_FACTORY_MAP.get(script_info.handler_utype)
    if unit_factory is None:
        eventlog.log_e(f'unsupported (handler-utype:{script_info.handler_utype})')
        raise InternalError

    try:
        firmware: PsuctlFirmware | VitekFirmware | None = None

        firmware_factory = _UTYPE_TO_FIRMWARE_FACTORY_MAP.get(script_info.handler_utype)
        if firmware_factory is not None:
            firmware = firmware_factory(endpoint)

    except GenericError:
        eventlog.log_e(f'fail: firmware_factory (handler-utype:{script_info.handler_utype})')
        raise InternalError

    unit = unit_factory(firmware, script_info.handler_uref)
    if unit is None:
        eventlog.log_e(f'fail: unit_factory (handler-uref:{script_info.handler_uref})')
        raise InternalError

    script_conf = script_info.handler_defconf
    script_conf.update(override_conf)

    if lock_path:
        with fs.LockContext(lock_path, try_count=3):
            unit.run_script(script_conf)
    else:
        unit.run_script(script_conf)


class Lens:
    def __init__(self, uref: str) -> None:
        psuctl_uconf = PsuctlUconf(uref)
        iris_script_info = psuctl_uconf.get_script_info('rotate-iris')
        focus_script_info = psuctl_uconf.get_script_info('rotate-focus')

        if iris_script_info.handler_utype != focus_script_info.handler_utype:
            eventlog.log_e(
                'unsupported (iris-handler-utype:{}, focus-handler-utype:{})'
                .format(iris_script_info.handler_utype, focus_script_info.handler_utype))
            raise InternalError

        handler_utype: Final[str] = iris_script_info.handler_utype

        if handler_utype not in _UTYPE_TO_FIRMWARE_FACTORY_MAP:
            eventlog.log_e(f'unsupported (handler-utype:{handler_utype})')
            raise InternalError

        unit_factory = _UTYPE_TO_UNIT_FACTORY_MAP.get(handler_utype)
        if unit_factory is None:
            eventlog.log_e(f'unsupported (handler-utype:{handler_utype})')
            raise InternalError

        # firmware
        try:
            firmware: PsuctlFirmware | VitekFirmware | None = None

            firmware_factory = _UTYPE_TO_FIRMWARE_FACTORY_MAP.get(handler_utype)
            if firmware_factory is not None:
                firmware = firmware_factory(psuctl_uconf.endpoint)

        except GenericError:
            eventlog.log_e(f'fail: firmware_factory (handler-utype:{handler_utype})')
            raise InternalError

        self._iris_unit = unit_factory(firmware, iris_script_info.handler_uref)
        self._focus_unit = unit_factory(firmware, focus_script_info.handler_uref)

        self._iris_rotate_defconf = iris_script_info.handler_defconf
        self._focus_rotate_defconf = focus_script_info.handler_defconf

        self._lock = fs.LockContext(psuctl_uconf.lock_path, try_count=3)

        with self._lock:
            script_conf = self._iris_rotate_defconf.copy()
            script_conf.update({'action': 'get-limits'})
            self.iris_limits: Final[tuple[int, int]] = self._iris_unit.run_script(script_conf)

            script_conf = self._focus_rotate_defconf.copy()
            script_conf.update({'action': 'get-limits'})
            self.focus_limits: Final[tuple[int, int]] = self._focus_unit.run_script(script_conf)

    def __enter__(self):
        self._lock.__enter__()
        return self

    def __exit__(self, x_type, x_self, x_traceback):
        self._lock.__exit__(x_type, x_self, x_traceback)

    def rotate_iris(self, new_iris: int, old_iris: int) -> None:
        min_iris, max_iris = self.iris_limits

        if old_iris != new_iris:
            if new_iris == 0:
                iris_steps = -(max_iris - min_iris)
            else:
                iris_steps = int(new_iris - old_iris)

            script_conf = self._iris_rotate_defconf.copy()
            script_conf.update({'steps': iris_steps})

            self._iris_unit.run_script(script_conf)

    def rotate_focus(self, new_focus: int, old_focus: int) -> None:
        min_focus, max_focus = self.focus_limits
        if old_focus != new_focus:
            if new_focus == 0:
                focus_steps = -(max_focus - min_focus)
            else:
                focus_steps = int(new_focus - old_focus)

            script_conf = self._focus_rotate_defconf.copy()
            script_conf.update({'steps': focus_steps})

            self._focus_unit.run_script(script_conf)
