import time

import serial

from edge import eventlog
from edge.types.exceptions import InternalError


class VitekFirmware:
    STAT_CHAN_STEPS = '1'

    CAP_CHAN_MAX_STEPS = '1'
    CAP_CHAN_STEP_TOUT = '2'

    REF_CHAN1 = 'ch1'
    REF_CHAN2 = 'ch2'
    REF_CHAN3 = 'ch3'

    def __init__(self, endpoint: str) -> None:
        # if pass endpoint to constructor it will try to open port and may throw exeption
        self._port = serial.Serial(port=endpoint, baudrate=9600, parity=serial.PARITY_NONE, stopbits=1)
        # self.ser.port(endpoint)
        # self._port.close()

        self._caps: dict[str, dict[str, int]] | None = None
        self._stat: dict[str, dict[str, int]] | None = None

    def tx_query(self, query_msg: str) -> int:
        try:
            self._port.reset_input_buffer()
            return self._port.write(query_msg)

        finally:
            self._stat = None

    def rx_reply(self) -> str:
        reply_msg = self._port.read_all()
        return reply_msg

    def _refresh(self) -> None:
        self._port.write('?')
        time.sleep(1.5)
        reply = self._port.read_all()

        def get(param: str) -> int | None:
            param_pos = reply.find(param)
            if -1 == param_pos:
                eventlog.log_e(f'fail: find start (param:{param})')
                raise InternalError

            param_pos += len(param)
            endlpos = reply.find('\n', param_pos)
            if -1 == endlpos:
                eventlog.log_e(f'fail: find end (param:{param})')
                raise InternalError

            # skip ":" after param-name
            param_value = reply[param_pos+1:endlpos]
            return int(param_value)

        ch1_now = get('Ch1 steps now')
        ch1_max = get('Full range Ch1 steps')
        ch2_now = get('Ch2 steps now')
        ch2_max = get('Full range Ch2 steps')
        ch3_now = get('Ch3 steps now')
        ch3_max = get('Full range Ch3 steps')
        step_tout1 = get('Stepper motor Ch1 period, x100us')
        step_tout2 = get('Stepper motor Ch2 period, x100us')
        step_tout3 = get('Stepper motor Ch3 period, x100us')
        if None in [ch1_now, ch1_max, ch2_now, ch3_max, ch3_now, ch2_max, step_tout1, step_tout2, step_tout3]:
            raise InternalError

        self._caps = {
            self.REF_CHAN1: {
                self.CAP_CHAN_MAX_STEPS: ch1_max,
                self.CAP_CHAN_STEP_TOUT: step_tout1
            },
            self.REF_CHAN2: {
                self.CAP_CHAN_MAX_STEPS: ch2_max,
                self.CAP_CHAN_STEP_TOUT: step_tout2
            },
            self.REF_CHAN3: {
                self.CAP_CHAN_MAX_STEPS: ch3_max,
                self.CAP_CHAN_STEP_TOUT: step_tout3
            }
        }

        self._stat = {
            self.REF_CHAN1: {
                self.STAT_CHAN_STEPS: ch1_now
            },
            self.REF_CHAN2: {
                self.STAT_CHAN_STEPS: ch2_now
            },
            self.REF_CHAN3: {
                self.STAT_CHAN_STEPS: ch3_now
            }
        }

    def get_chan_cap(self, chan_ref, cap_name) -> int:
        if self._caps is None:
            self._refresh()

        return self._caps[chan_ref][cap_name]

    def get_chan_stat(self, chan_ref, stat_name) -> int:
        if self._stat is None:
            self._refresh()

        return self._stat[chan_ref][stat_name]
