from typing import Any, Final

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import NotFoundError, ParamsError


class ScriptInfo:
    def __init__(self, handler_uref: tuple[str, str], handler_defconf: dict[str, Any]) -> None:
        """
        :param handler_uref:
            The reference of the psuctl unit.
        :param handler_defconf:
            The default configuration of the psuctl unit script.
        """
        self.handler_uref: Final[tuple[str, str]] = handler_uref
        self.handler_defconf: Final[dict[str, Any]] = handler_defconf

    @property
    def handler_utype(self) -> str:
        return self.handler_uref[0]


class ScriptsConf:
    @staticmethod
    def decode_script_info_fromjnode(scripts_jnode: dict[str, Any], script_name: str) -> ScriptInfo:
        """
        :param scripts_jnode:
            scripts#/scripts
        :param script_name:
            'led-camera' -> scripts#/scripts/led-camera

        :raises NotFoundError:
        """
        script_jnode: dict[str, Any] | None = scripts_jnode.get(script_name)
        if script_jnode is None:
            eventlog.log_e_param(f'#/{script_name}')
            raise NotFoundError()

        unit_type_ref = script_jnode.get('unit')
        if type(unit_type_ref) not in (list,):
            eventlog.log_e_param(f'#/{script_name}/unit')
            raise NotFoundError()

        elif len(unit_type_ref) != 2:
            eventlog.log_e_param(f'#/{script_name}/unit')
            raise NotFoundError()

        return ScriptInfo(unit_type_ref, script_jnode.get('conf') or dict())

    @staticmethod
    def decode_script_info_fromfile(scripts_path: str, script_name: str) -> ScriptInfo:
        jnode = fs.load_json3(scripts_path, default=None)
        if jnode is None:
            eventlog.log_e(f'fail: load <{scripts_path}>')
            raise NotFoundError()

        scripts_jnode: dict[str, Any] | None = jnode.get('scripts')
        if scripts_jnode is None:
            eventlog.log_e_param(f'<{scripts_path}#/scripts>')
            raise ParamsError()

        return ScriptsConf.decode_script_info_fromjnode(scripts_jnode, script_name)


class PsuctlUconf:
    def __init__(self, uref: str) -> None:
        """
        :param uref:
            'psuctl.0' -> /etc/edge/setupstore/admin/actions/psuctl.0.json

        :raises NotFoundError:
        :raises ParamsError:
        """
        jnode_path = f'/etc/edge/setupstore/admin/actions/{uref}.json'
        jnode = fs.load_json3(jnode_path, default=None)
        if jnode is None:
            eventlog.log_e(f'fail: load <{jnode_path}>')
            raise NotFoundError()

        uiconf = jnode.get(uref)
        if uiconf is None:
            eventlog.log_e_param(f'<{jnode_path}#/{uref}>')
            raise ParamsError()

        uiconf_pri = uiconf.get('pri')
        if uiconf_pri is None:
            eventlog.log_e_param(f'<{jnode_path}#/{uref}/pri>')
            raise ParamsError()

        # /pri/endpoint
        endpoint = uiconf_pri.get('endpoint')
        if endpoint is None:
            eventlog.log_e_param(f'<{jnode_path}#/{uref}/pri/endpoint>')
            raise ParamsError()

        # /pri/lock-path
        lock_path: str | None = uiconf_pri.get('lock-path')
        if lock_path is None:
            eventlog.log_e_param(f'<{jnode_path}#/{uref}/pri/lock-path>')
            raise ParamsError()

        elif type(lock_path) not in (str,):
            eventlog.log_e_param(f'<{jnode_path}#/{uref}/pri/lock-path>')
            raise ParamsError()

        # /pri/scripts
        scripts_value: dict[str, Any] | str | None = uiconf_pri.get('scripts')
        if type(scripts_value) in (str,):
            scripts_jnode = fs.load_json3(scripts_value, default=None)
            if scripts_jnode is None:
                eventlog.log_e(f'fail: load <{scripts_value}>')
                raise NotFoundError()

            scripts_jnode: dict[str, Any] = scripts_jnode.get('scripts')
            if scripts_jnode is None:
                eventlog.log_e_param(f'<{scripts_value}#/scripts>')
                raise ParamsError()

        elif type(scripts_value) in (dict,):
            scripts_jnode: dict[str, Any] = scripts_value

        else:
            eventlog.log_e_param(f'<{jnode_path}#/{uref}/pri/scripts>')
            raise ParamsError()

        self.endpoint: Final[str] = endpoint
        self.lock_path: Final[str] = lock_path
        self._scripts_jnode: Final[dict[str, Any]] = scripts_jnode

    def get_script_info(self, script_name: str) -> ScriptInfo:
        """
        :param script_name:
            'led-camera' -> (/etc/edge/setupstore/admin/actions/psuctl.0.json#/psuctl.0/pri/scripts)#/scripts/led-camera

        :raises NotFoundError:
        :raises ParamsError:
        """
        return ScriptsConf.decode_script_info_fromjnode(self._scripts_jnode, script_name)
