__all__ = ['factory']

import time

from edge import eventlog

from edge.types.exceptions import InternalError, GenericError
from psuctl.firmware.psuctl import PsuctlFirmware
from psuctl.firmware.vitek import VitekFirmware
from psuctl.types.unit import IUnit, UnitRef, Unit

_ACK_TOUT_SEC = 2.0


class _PsuUnit10(Unit[PsuctlFirmware]):
    _REF_TO_OPBIT_MAP: dict[UnitRef, int] = {
        ('psu', 'BASE_V1'): 0x0000,
        ('psu', 'BASE_V2'): 0x0001,
        ('psu', 'BASE_V3'): 0x0002,
        ('psu', 'BASE_V4'): 0x0003,
        ('psu', 'BASE_USB1'): 0x000C,
        ('psu', 'BASE_USB2'): 0x000D
    }

    def __init__(self, firmware: PsuctlFirmware, ref: UnitRef) -> None:
        if tuple(ref) not in self._REF_TO_OPBIT_MAP:
            eventlog.log_e(f'fail: ivalid ref <{ref}>')
            raise InternalError

        super().__init__(firmware, ref)

    def _write_bit(self, op_bit: int, op_val: bool) -> None:
        try:
            answer = self._firmware.write_rw_bit(op_bit, op_val, _ACK_TOUT_SEC)
            if answer != op_val:
                eventlog.log_w(f'fail: write_rw_bit (query:{op_val}, reply:{answer})')

        except InternalError:
            eventlog.log_e('fail: write_rw_reg')
            raise

    def run_script(self, conf: dict) -> tuple[int, int] | None:
        """
        :param conf:
            psuctl.conf#/scripts/<name>/conf
        """
        action = conf.get('action')
        if action not in ['on', 'off', 'reboot']:
            eventlog.log_e(f'fail: unsupported action <{action}>')
            raise InternalError

        op_bit = self._REF_TO_OPBIT_MAP[tuple(self._ref)]
        if action == 'reboot':
            self._write_bit(op_bit, False)
            time.sleep(1)
            self._write_bit(op_bit, True)

        else:
            self._write_bit(op_bit, action == 'on')

        return None


class _PsuUnit14(Unit):
    _REF_TO_INFO_MAP: dict[UnitRef, int] = {
        ('psu', 'BASE_V1'): 0x1000,
        ('psu', 'BASE_V2'): 0x1001,
        ('psu', 'BASE_V3'): 0x1002,
        ('psu', 'BASE_V4'): 0x1003,
        ('psu', 'BASE_USB1'): 0x100C,
        ('psu', 'BASE_USB2'): 0x100D
    }

    _ACTION_MAP: dict[str, int] = {
        'on': 0xFF00,
        'off': 0x0000,
        'reboot': 0x0002
    }

    def __init__(self, firmware: PsuctlFirmware, ref: UnitRef) -> None:
        if tuple(ref) not in self._REF_TO_INFO_MAP:
            eventlog.log_e(f'fail: ivalid ref <{ref}>')
            raise InternalError

        super().__init__(firmware, ref)

    def run_script(self, conf: dict) -> tuple[int, int] | None:
        """
        :param conf:
            psuctl.conf#/scripts/<name>/conf
        """

        action = conf.get('action')
        if action not in ['on', 'off', 'reboot']:
            eventlog.log_e(f'fail: unsupported action <{action}>')
            raise InternalError

        reg_address = self._REF_TO_INFO_MAP[tuple(self._ref)]
        try:
            if action == 'reboot':
                reg_value = self._ACTION_MAP['off']
                answer = self._firmware.write_rw_reg(reg_address, reg_value, _ACK_TOUT_SEC)
                if reg_value != answer:
                    eventlog.log_w(f'fail: write_rw_reg (query:{reg_value}, reply:{answer})')

                reg_value = self._ACTION_MAP['on']
                answer = self._firmware.write_rw_reg(reg_address, reg_value, _ACK_TOUT_SEC)
                if reg_value != answer:
                    eventlog.log_w(f'fail: write_rw_reg (query:{reg_value}, reply:{answer})')
            else:
                reg_value = self._ACTION_MAP[action]
                answer = self._firmware.write_rw_reg(reg_address, reg_value, _ACK_TOUT_SEC)
                if reg_value != answer:
                    eventlog.log_w(f'fail: write_rw_reg (query:{reg_value}, reply:{answer})')

        except GenericError:
            eventlog.log_e('fail: write_rw_reg')
            raise

        return None


def factory(firmware: PsuctlFirmware | VitekFirmware | None, unit_ref: UnitRef) -> IUnit:
    if type(firmware) is not PsuctlFirmware:
        eventlog.log_e('fail: unsupported firmware')
        raise InternalError

    firmware_ver = firmware.version
    if firmware_ver in ['v1.0', 'v1.1', 'v1.2', 'v1.3']:
        unit_factory = _PsuUnit10

    elif firmware_ver in ['v1.4', 'v1.5']:
        unit_factory = _PsuUnit14

    else:
        unit_factory = _PsuUnit14
        eventlog.log_w(f'fail: unsupported firmware <{firmware_ver}> (using {unit_factory.__name__})')

    return unit_factory(firmware, unit_ref)
