__all__ = ['factory']

import time

from edge import eventlog
from edge.types.exceptions import InternalError
from psuctl.firmware.psuctl import PsuctlFirmware
from psuctl.firmware.vitek import VitekFirmware
from psuctl.types.unit import IUnit, UnitRef, Unit


class _VitekStepperUnit(Unit[VitekFirmware]):
    _REF_TO_INFO_MAP: dict[UnitRef, dict[str, str]] = {
        ('stepper-vitek', 'ch1'): {
            'down': 'a',
            'up': 'A'
        },
        ('stepper-vitek', 'ch2'): {
            'down': 'b',
            'up': 'B'
        },
        ('stepper-vitek', 'ch3'): {
            'down': 'c',
            'up': 'C'
        }
    }

    _STEPS_PER_CLICK_TO_CMD: list[tuple[int, str]] = [
        (1000, '7'), (400, '6'), (100, '5'),
        (40, '4'), (10, '3'), (4, '2'), (1, '1')
    ]

    def __init__(self, firmware: VitekFirmware, ref: UnitRef) -> None:
        if tuple(ref) not in self._REF_TO_INFO_MAP:
            eventlog.log_e(f'fail: ivalid ref <{ref}>')
            raise InternalError

        super().__init__(firmware, ref)

        self._cap_step_tout = firmware.get_chan_cap(ref[1], firmware.CAP_CHAN_STEP_TOUT) * 100 * 1e-6
        self._cap_max_steps = firmware.get_chan_cap(ref[1], firmware.CAP_CHAN_MAX_STEPS)

    def _rotate(self, steps: int) -> None:
        rot_steps = abs(steps)
        if rot_steps > self._cap_max_steps:
            eventlog.log_e(f'fail: invalid steps value <{steps}>')
            raise InternalError

        rot_dir = 'up' if steps > 0 else 'down'
        rot_dir_op = self._REF_TO_INFO_MAP[tuple(self._ref)][rot_dir]

        # setup direction
        byte_count = self._firmware.tx_query(rot_dir_op)
        eventlog.log_d(f'done: setup dir (dir:{rot_dir}, op:{rot_dir_op}, tx-bytes:{byte_count})')

        time.sleep(0.010)
        for click_steps, click_op in self._STEPS_PER_CLICK_TO_CMD:
            if click_steps > rot_steps:
                continue

            # setup click-size
            byte_count = self._firmware.tx_query(click_op)
            eventlog.log_d(f'done: setup click (steps:{click_steps}, op:{click_op}, tx-bytes:{byte_count})')
            time.sleep(0.010)
            # rotate motor
            while click_steps <= rot_steps:
                byte_count = self._firmware.tx_query(' ')
                eventlog.log_d(f'done: rotate (steps:{click_steps}, tx-bytes:{byte_count})')

                click_tout = click_steps * self._cap_step_tout * 4
                time.sleep(click_tout)
                rot_steps -= click_steps

    def run_script(self, conf: dict) -> tuple[int, int] | None:
        # unpack #/action
        action = conf.get('action')

        if action not in ['rotate', 'reset', 'set', 'get-limits']:
            eventlog.log_e_param('#/action')
            raise InternalError

        def unpack_str_int(key: str) -> int:
            value = conf.get(key)
            if type(value) in [str]:
                return int(value)

            elif type(value) in [int]:
                return value

            else:
                eventlog.log_e_param(f'#/{key}')
                raise InternalError

        if action == 'get-limits':
            return 0, self._cap_max_steps

        elif action == 'rotate':
            steps = unpack_str_int('steps')
            self._rotate(steps)

        elif action == 'reset':
            position = unpack_str_int('position')
            steps_now = self._firmware.get_chan_stat(self._ref[1], self._firmware.STAT_CHAN_STEPS)

            self._rotate(-steps_now)
            time.sleep(0.100)
            self._rotate(position)

        elif action == 'set':
            position = unpack_str_int('position')
            steps_now = self._firmware.get_chan_stat(self._ref[1], self._firmware.STAT_CHAN_STEPS)
            self._rotate(position - steps_now)


def factory(firmware: PsuctlFirmware | VitekFirmware | None, unit_ref: UnitRef) -> IUnit:
    if type(firmware) is not VitekFirmware:
        eventlog.log_e('fail: unsupported firmware')
        raise InternalError

    return _VitekStepperUnit(firmware, unit_ref)
